/**************************************************************
SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
Copyright (c)2016 Chainik, SVP-Team - https://www.svp-team.com

Unified interface between core logic and native Avisynth,
Vapoursynth, etc interfaces.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#pragma once

#include "common.h"

class CoreLogic;

class BaseFilter
{
public:
    class Frame
    {
    public:
#ifdef AVISYNTH_PLUGIN
        Frame(PVideoFrame frm, int bpp);
#endif
#ifdef VAPOURSYNTH_PLUGIN
        Frame(VSFrameRef *frm, const VSAPI *vsapi);
        Frame(const VSFrameRef *frm, const VSAPI *vsapi);

		void getDuration(int64_t& num, int64_t& den) const;
		void setDuration(int64_t num, int64_t den);

        int64_t getPTS() const;
        void setPTS(int64_t pts);
#endif
        virtual ~Frame();

        bool isNull() const;
        void setReleaseOnDelete(bool on);

        int width() const;
        int height() const;
        int pitchY() const;
        int pitchUV() const;

        bool read (uint8_t** planeY,uint8_t** planeU=0,uint8_t** planeV=0) const;
        bool write(uint8_t** planeY,uint8_t** planeU=0,uint8_t** planeV=0);

    protected:
#ifdef AVISYNTH_PLUGIN
        PVideoFrame avsFrame;
        int bpp;
        friend class AVSBaseFilter;
#endif
#ifdef VAPOURSYNTH_PLUGIN
        bool releaseOnDelete;
        VSFrameRef *vsFrame;
        bool writeable;
        const VSAPI *vsapi;
        int64_t durNum, durDen, pts;
        friend class VSBaseFilter;
#endif
    };

    class Video
    {
    public:
        enum PixelType
        {
            PixelUnknown,
            PixelData,
            PixelVideo_YV12 =10,
            PixelVideo_YUV_P10,
            PixelVideo_YUV_P16
        };

protected:
#ifdef AVISYNTH_PLUGIN
        Video(int w,int h,PixelType type);
        Video(const VideoInfo& vi);
#endif
#ifdef VAPOURSYNTH_PLUGIN
        Video(int w,int h,PixelType type, const VSAPI *vsapi, VSCore *vscore);
        Video(const VSVideoInfo &vi, const VSAPI *vsapi, VSCore *vscore);
#endif

public:
        int width() const;
        int height() const;
        void setSize(int w,int h);

        void frameRate(uint32_t &num, uint32_t &den) const;
        void setFrameRate(uint32_t num,uint32_t den);

        PixelType pixelType() const;
        void setPixelType(PixelType type);

        uint32_t totalFrames() const;
        void setTotalFrames(uint32_t count);

    protected:
#ifdef AVISYNTH_PLUGIN
        VideoInfo avs_vi;
        friend class AVSBaseFilter;
#endif
#ifdef VAPOURSYNTH_PLUGIN
        VSVideoInfo vs_vi;
        const VSAPI *vsapi;
        VSCore *vscore;
        friend class VSBaseFilter;
#endif
    };

public:
    BaseFilter();
    virtual ~BaseFilter();

    void setCore(CoreLogic *core);

    std::string getError() const;
    void setError(const std::string& error);

    virtual void setInt64(const std::string& name,int64_t i) =0;
    virtual int64_t getInt64(const std::string& name) =0;

    Video* getVideoInfo() const;
    virtual Video getVideoInfo(const std::string& name) const =0;
    virtual Video newVideoInfo(int w,int h,Video::PixelType type) const =0;

    virtual bool hasUpstreamFilter(const std::string& name) const =0;

    virtual std::shared_ptr<Frame> newFrame(void *baseRef,std::shared_ptr<Frame> src=std::shared_ptr<Frame>(0)) =0;
    virtual std::shared_ptr<Frame> newFrame(const Video& vi,void *baseRef,std::shared_ptr<Frame> src=std::shared_ptr<Frame>(0)) =0;
#ifdef VAPOURSYNTH_PLUGIN
    virtual std::shared_ptr<Frame> copyFrame(void *baseRef,std::shared_ptr<Frame> src) const =0;
#endif

    virtual void prefetchUpstreamFrame(const std::string& filter,int n,void *baseRef) =0;
    virtual std::shared_ptr<Frame> getUpstreamFrame(const std::string& filter,int n,void *baseRef) =0;

protected:
    CoreLogic* getCore();

protected:
    Video* videoInfo;

private:
    CoreLogic* core;
    std::string errorString;
};

class CoreLogic
{
public:
    CoreLogic(BaseFilter *base);
    virtual ~CoreLogic();

    virtual void initVideoInfo(BaseFilter::Video *vi) =0;

    virtual void prepare(int n,void* baseRef) =0;
    virtual std::shared_ptr<BaseFilter::Frame> get(int n,void* baseRef) =0;

    BaseFilter* getBase() const;

private:
    BaseFilter *base;
};

#ifdef AVISYNTH_PLUGIN
class AVSBaseFilter:public BaseFilter
{
public:
    AVSBaseFilter(const VideoInfo& vi,IScriptEnvironment* env);
    virtual ~AVSBaseFilter();

    //Native interface
    PVideoFrame getFrame(int n,IScriptEnvironment *env);

    void addUpstreamFilter(const std::string& name,PClip clip);
    bool hasUpstreamFilter(const std::string& name) const;

    Video getVideoInfo(const std::string& name) const;
    Video newVideoInfo(int w,int h,Video::PixelType type) const;

    VideoInfo initVideoInfo();

    //BaseFilter interface
    std::shared_ptr<Frame> newFrame(void* baseRef,std::shared_ptr<Frame> src=std::shared_ptr<Frame>(0));
    std::shared_ptr<Frame> newFrame(const Video& vi,void* baseRef,std::shared_ptr<Frame> src=std::shared_ptr<Frame>(0));

    void prefetchUpstreamFrame(const std::string& /*filter*/,int /*n*/,void* /*baseRef*/) {}
    std::shared_ptr<Frame> getUpstreamFrame(const std::string& filter,int n,void* baseRef);

    void setInt64(const std::string& name,int64_t i);
    int64_t getInt64(const std::string& name);

protected:
    IScriptEnvironment* mainEnv;
    std::map<std::string,PClip> clips;
    bool vi_inited;
    bool avsV8;
};
#endif

#ifdef VAPOURSYNTH_PLUGIN
class VSBaseFilter:public BaseFilter
{
public:
    VSBaseFilter(const VSVideoInfo &vi, VSCore *vscore, const VSAPI *vsapi);
    virtual ~VSBaseFilter();

    //Native interface
    void prepareFrame(int n,VSFrameContext *frameCtx);
    VSFrameRef *getFrame(int n,VSFrameContext *frameCtx);

    void addUpstreamFilter(const std::string& name,VSNodeRef *node);
    bool hasUpstreamFilter(const std::string& name) const;

    Video getVideoInfo(const std::string& name) const;
    Video newVideoInfo(int w,int h,Video::PixelType type) const;

    VSVideoInfo initVideoInfo();

    //BaseFilter interface
    std::shared_ptr<Frame> newFrame(void* /*baseRef*/,std::shared_ptr<Frame> src=std::shared_ptr<Frame>(0));
    std::shared_ptr<Frame> newFrame(const Video& vi,void* /*baseRef*/,std::shared_ptr<Frame> src=std::shared_ptr<Frame>(0));
    std::shared_ptr<Frame> copyFrame(void* /*baseRef*/,std::shared_ptr<Frame> src) const;

    void prefetchUpstreamFrame(const std::string& filter,int n,void* baseRef);
    std::shared_ptr<Frame> getUpstreamFrame(const std::string& filter,int n,void* baseRef);

    void setInt64(const std::string& name,int64_t i);
    int64_t getInt64(const std::string& name);

protected:
    const VSAPI *vsapi;
    VSCore *vscore;
    std::map<std::string,VSNodeRef*> nodes;
    std::map<std::string,int64_t> nodesParams;
};
#endif
