/**************************************************************
SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
Copyright(c)2016 Chainik - http://www.svp-team.com

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#include <algorithm>

#include "vectorsconverter.h"
#include "groupofplanes.h"

#define MOTION_MAGIC_KEY			0x564D
#define MOTION_USE_ISSE             0x00000020
#define MOTION_IS_BACKWARD          0x00000040
#define MOTION_USE_CHROMA_MOTION    0x00000800

#ifdef AVISYNTH_PLUGIN

VectorsConverter::VectorsConverter(PClip vectors,bool isb,IScriptEnvironment* env):
    GenericVideoFilter(vectors)
{
    if (vi.nchannels != 0 || vi.audio_samples_per_second != 0)
        env->ThrowError("SVConvert: invalid vectors stream [1]");

    SVAnalysisData *data = reinterpret_cast<SVAnalysisData*>(vi.num_audio_samples);
    if(data->nVersion != MVANALYSIS_DATA_VERSION)
        env->ThrowError("SVConvert: invalid vectors stream [2]");

    if((isb && !(data->vectorsType&2)) || (!isb && !(data->vectorsType&1)))
        env->ThrowError("SVConvert: SVAnalyse clip doesn't contains %s vectors",isb?"backward":"forward");

    ad.isBackward=isb;
    ad.nBlkSizeX=data->nBlkSizeX;
    ad.nBlkSizeY=data->nBlkSizeY;
    ad.nBlkX=data->nBlkX;
    ad.nBlkY=data->nBlkY;
    ad.nDeltaFrame=data->delta;
    ad.nFlags=0;
    ad.nHeight=data->nHeight;
    ad.nHPadding=0;
    ad.nLvCount=1;
    ad.nMagicKey=MOTION_MAGIC_KEY;
    ad.nOverlapX=data->nOverlapX;
    ad.nOverlapY=data->nOverlapY;
    ad.nPel=data->nPel;
    ad.nVersion=5; //old MVTools DATA_VERSION
    ad.nVPadding=0;
    ad.nWidth=data->nWidth;
    ad.pixelType=VideoInfo::CS_YV12;
    ad.xRatioUV=2;
    ad.yRatioUV=2;

    ad.nFlags |= MOTION_USE_ISSE|MOTION_USE_CHROMA_MOTION;
    ad.nFlags |= ad.isBackward ? MOTION_IS_BACKWARD : 0;

    headerSize = _fmax(4 + (unsigned int)sizeof(ad), 256u);

    vi.height = 1;
    vi.width = headerSize/sizeof(int)+2+1+ad.nBlkX*ad.nBlkY*3;
    vi.pixel_type = VideoInfo::CS_BGR32;
	vi.audio_samples_per_second = 0;
#ifndef Q_PROCESSOR_X86_64
    vi.nchannels = (int)&ad;
#else
	uintptr_t p = reinterpret_cast <uintptr_t> (&ad);
	vi.nchannels = 0x80000000L | (int)(p >> 32);
	vi.sample_type = (int)(p & 0xffffffffUL);
#endif
}

VectorsConverter::~VectorsConverter()
{
}

PVideoFrame __stdcall VectorsConverter::GetFrame(int n, IScriptEnvironment* env)
{
    PVideoFrame fn=child->GetFrame(ad.isBackward ? n : _fmax(n-ad.nDeltaFrame,0),env);

    const int *data = reinterpret_cast<const int*>(fn->GetReadPtr());
    if (data[0] != 1 + sizeof(SVAnalysisData) / sizeof(int) || data[1] != MVANALYSIS_DATA_VERSION)
        return 0;

    bool valid=(!ad.isBackward ? (n-ad.nDeltaFrame>=0) : (data[2]?true:false));
    data+=data[0];

    PVideoFrame	dst=env->NewVideoFrame(vi);
    int *pDst = (int*)(dst->GetWritePtr());

    pDst[0]=headerSize;
    memcpy(pDst+1,&ad,sizeof(ad));
    pDst += headerSize/sizeof(int);

    int sz=ad.nBlkX*ad.nBlkY;
    if(ad.isBackward)
        data+=1+sz*2;

    *(pDst++)=2+1+sz*3;
    *(pDst++)=(valid?1:0);
    *(pDst++)=1+sz*3;
    if(valid)
    {
        if(*(data++) != 1+sz*2) return 0;

        //shift=0 -> *1/1, 1 -> *3/4, 2 -> *1/2
        int chromaMul=(CHROMA_SHIFT==1)?3:1,chromaDiv=(CHROMA_SHIFT==0?0:(CHROMA_SHIFT==1?2:1));
        for(int by=0, py=0; by<ad.nBlkY; by++, py+=ad.nBlkSizeY-ad.nOverlapY)
            for(int bx=0, px=0; bx<ad.nBlkX; bx++, px+=ad.nBlkSizeX-ad.nOverlapX, data+=2)
        {
            short x = (data[0] >> 16), y = (data[0] & 0x00ffff);
            int sad = (data[1] & 0x00ffffff);

            x = _fmax(-px, _fmin(ad.nWidth - px - ad.nBlkSizeX, (int)x));
            y = _fmax(-py, _fmin(ad.nHeight - py - ad.nBlkSizeY, (int)y));

            *(pDst++)=x;
            *(pDst++)=y;
            *(pDst++)=(sad*chromaMul)>>chromaDiv; //SAD
        }
    }
    return dst;
}

#endif
